# Copyright 2024 The AI Edge Torch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Provides a configuration for the ai-edge-torch."""

import functools
import logging
import os

__all__ = ["config"]


def _get_bool_env_var(name: str, default: bool) -> bool:
  var = os.environ.get(name, "false")
  var = var.lower().strip()
  if var in ("y", "yes", "t", "true", "on", "1"):
    return True
  elif var in ("n", "no", "f", "false", "off", "0"):
    return False
  else:
    logging.warning("Invalid %s value is ignored: %s.", name, var)
    return default


class _Config:
  """ai-edge-torch global configs."""

  @property
  @functools.cache  # pylint: disable=method-cache-max-size-none
  def use_torch_xla(self) -> bool:
    """True if using torch_xla to lower torch ops to StableHLO.

    To use torch_xla as the lowering backend, set environment variable
    `USE_TORCH_XLA` to "true".
    """
    return _get_bool_env_var("USE_TORCH_XLA", default=False)

  @property
  def in_oss(self) -> bool:
    """True if the code is not running in google internal environment."""
    return True

  @property
  def enable_group_norm_composite(self) -> bool:
    """True if lowering group norm in StableHLO composite.

    Currently only supports NHWC group norm generated by
    OptimizeLayoutTransposesPass.
    """
    return _get_bool_env_var("ENABLE_GROUP_NORM_COMPOSITE", default=False)

  @enable_group_norm_composite.setter
  def enable_group_norm_composite(self, value: bool):
    os.environ["ENABLE_GROUP_NORM_COMPOSITE"] = "y" if value else "n"

  @property
  def layout_optimize_partitioner(self) -> str:
    """The algorithm to use for layout optimization."""
    return os.environ.get("AIEDGETORCH_LAYOUT_OPTIMIZE_PARTITIONER", "DEFAULT")

  @layout_optimize_partitioner.setter
  def layout_optimize_partitioner(self, value: str):
    os.environ["AIEDGETORCH_LAYOUT_OPTIMIZE_PARTITIONER"] = str(value).upper()


config = _Config()
